@@ -1,10 +1,10 @@
 ##----------------------------------------------------------------------------
 ## JSON Schema Validator - ~/lib/JSON/Schema/Validate.pm
-## Version v0.2.0
+## Version v0.3.0
 ## Copyright(c) 2025 DEGUEST Pte. Ltd.
 ## Author: Jacques Deguest <jack@deguest.jp>
 ## Created 2025/11/07
-## Modified 2025/11/09
+## Modified 2025/11/10
 ## All rights reserved
 ## 
 ## 
@@ -23,7 +23,7 @@ BEGIN
     use Scalar::Util qw( blessed looks_like_number reftype refaddr );
     use List::Util qw( first any all );
     use Encode ();
-    our $VERSION = 'v0.2.0';
+    our $VERSION = 'v0.3.0';
 };
 
 use strict;
@@ -710,7 +710,7 @@ sub _apply_ref
     {
         local $@;
         my $ext = eval{ $ctx->{resolver}->( $abs ) };
-        return( _err_res( $ctx, $schema_ptr, "resolver failed for \$ref: $abs" ) ) unless( $ext && ref( $ext ) );
+        return( _err_res( $ctx, $schema_ptr, "resolver failed for \$ref: $abs", '$ref' ) ) unless( $ext && ref( $ext ) );
 
         my $ext_base = _normalize_uri( ( ref( $ext ) eq 'HASH' && $ext->{'$id'} ) ? $ext->{'$id'} : $abs );
 
@@ -736,7 +736,7 @@ sub _apply_ref
                 else
                 {
                     pop( @{$ctx->{id_stack}} );
-                    return( _err_res( $ctx, $schema_ptr, "unresolved JSON Pointer fragment in \$ref: $abs" ) );
+                    return( _err_res( $ctx, $schema_ptr, "unresolved JSON Pointer fragment in \$ref: $abs", '$ref' ) );
                 }
             }
             elsif( $frag =~ /\A\#([A-Za-z0-9._-]+)\z/ )
@@ -757,7 +757,7 @@ sub _apply_ref
         return( $r );
     }
 
-    return( _err_res( $ctx, $schema_ptr, "unresolved \$ref: $ref (abs: $abs)" ) );
+    return( _err_res( $ctx, $schema_ptr, "unresolved \$ref: $ref (abs: $abs)", '$ref' ) );
 }
 
 sub _canon
@@ -957,7 +957,7 @@ sub _compile_node
                     my $dec = _content_decode( $ctx, $S->{contentEncoding}, $bytes );
                     if( !defined( $dec ) )
                     {
-                        return( _err_res( $ctx, $ptr, "contentEncoding '$S->{contentEncoding}' decode failed" ) ) if( $assert );
+                        return( _err_res( $ctx, $ptr, "contentEncoding '$S->{contentEncoding}' decode failed", 'contentEncoding' ) ) if( $assert );
                     }
                     else
                     {
@@ -973,7 +973,7 @@ sub _compile_node
                         my( $ok, $msg, $maybe_decoded ) = _safe_invoke( $cb, $bytes, $params );
                         if( !$ok )
                         {
-                            return( _err_res( $ctx, $ptr, ( $msg || "contentMediaType '$mt' validation failed" ) ) ) if( $assert );
+                            return( _err_res( $ctx, $ptr, ( $msg || "contentMediaType '$mt' validation failed" ), 'contentMediaType' ) ) if( $assert );
                         }
                         # If the media validator decoded into a Perl structure, keep it
                         $decoded_ref = $maybe_decoded if( ref( $maybe_decoded ) );
@@ -993,7 +993,7 @@ sub _compile_node
                             } ? 1 : 0;
                             if( !$ok && $assert )
                             {
-                                return( _err_res( $ctx, $ptr, "contentMediaType '$mt' invalid UTF-8" ) );
+                                return( _err_res( $ctx, $ptr, "contentMediaType '$mt' invalid UTF-8", 'contentMediaType' ) );
                             }
                         }
                     }
@@ -1017,7 +1017,7 @@ sub _compile_node
 
                     if( !defined( $val ) )
                     {
-                        return( _err_res( $ctx, $ptr, "contentSchema present but payload not JSON-decodable" ) ) if( $assert );
+                        return( _err_res( $ctx, $ptr, "contentSchema present but payload not JSON-decodable", 'contentSchema' ) ) if( $assert );
                     }
                     else
                     {
@@ -1073,7 +1073,7 @@ sub _compile_node
                     last;
                 }
             }
-            return( _err_res( $ctx, $ptr, "instance does not satisfy anyOf" ) ) unless( $ok );
+            return( _err_res( $ctx, $ptr, "instance does not satisfy anyOf", 'anyOf' ) ) unless( $ok );
         }
         if( exists( $S->{oneOf} ) && ref( $S->{oneOf} ) eq 'ARRAY' )
         {
@@ -1083,12 +1083,12 @@ sub _compile_node
                 my $r = $child{"oneOf:$i"}->( $ctx, $inst );
                 $hits++ if( $r->{ok} );
             }
-            return( _err_res( $ctx, $ptr, "instance satisfies $hits schemas in oneOf (expected exactly 1)" ) ) unless( $hits == 1 );
+            return( _err_res( $ctx, $ptr, "instance satisfies $hits schemas in oneOf (expected exactly 1)", 'oneOf' ) ) unless( $hits == 1 );
         }
         if( exists( $S->{not} ) && ref( $S->{not} ) )
         {
             my $r = $child{"not"}->( $ctx, $inst );
-            return( _err_res( $ctx, $ptr, "instance matches forbidden not-schema" ) ) if( $r->{ok} );
+            return( _err_res( $ctx, $ptr, "instance matches forbidden not-schema", 'oneOf' ) ) if( $r->{ok} );
         }
 
         # Conditionals
@@ -1169,17 +1169,28 @@ sub _content_decode
 # Errors, utils, pointers, URIs
 sub _err
 {
-    my( $ctx, $ptr, $msg ) = @_;
+    my( $ctx, $schema_ptr, $msg, $keyword ) = @_;
     return(0) if( $ctx->{error_count} >= $ctx->{max_errors} );
-    push( @{$ctx->{errors}}, JSON::Schema::Validate::Error->new( $ptr, $msg ) );
+
+    # Instance path: use current ptr_stack top if available, else '#'
+    my $inst_path = $ctx->{ptr_stack} && @{$ctx->{ptr_stack}}
+                  ? ( $ctx->{ptr_stack}->[-1] // '#' )
+                  : '#';
+
+    push( @{$ctx->{errors}}, JSON::Schema::Validate::Error->new(
+            path            => $inst_path,
+            message         => $msg,
+            keyword         => $keyword,        # may be undef (back-compat)
+            schema_pointer  => $schema_ptr,     # where in the schema this came from
+    ));
     $ctx->{error_count}++;
     return(0);
 }
 
 sub _err_res
 {
-    my( $ctx, $ptr, $msg ) = @_;
-    _err( $ctx, $ptr, $msg );
+    my( $ctx, $schema_ptr, $msg, $keyword ) = @_;
+    _err( $ctx, $schema_ptr, $msg, $keyword );
     return( { ok => 0, props => {}, items => {} } );
 }
 
@@ -1316,11 +1327,11 @@ sub _k_array_all
 
     if( exists( $S->{minItems} ) && @$A < $S->{minItems} )
     {
-        return( _err_res( $ctx, $sp, "array has fewer than minItems $S->{minItems}" ) );
+        return( _err_res( $ctx, $sp, "array has fewer than minItems $S->{minItems}", 'minItems' ) );
     }
     if( exists( $S->{maxItems} ) && @$A > $S->{maxItems} )
     {
-        return( _err_res( $ctx, $sp, "array has more than maxItems $S->{maxItems}" ) );
+        return( _err_res( $ctx, $sp, "array has more than maxItems $S->{maxItems}", 'maxItems' ) );
     }
 
     if( $S->{uniqueItems} )
@@ -1331,7 +1342,7 @@ sub _k_array_all
             my $k = _canon( $A->[$i] );
             if( $seen{ $k }++ )
             {
-                return( _err_res( $ctx, _join_ptr( $sp, $i ), "array items not unique" ) );
+                return( _err_res( $ctx, _join_ptr( $sp, $i ), "array items not unique", 'uniqueItems' ) );
             }
         }
     }
@@ -1386,8 +1397,8 @@ sub _k_array_all
         my $minc = defined( $S->{minContains} ) ? $S->{minContains} : 1;
         my $maxc = defined( $S->{maxContains} ) ? $S->{maxContains} : ( 2**31 - 1 );
 
-        return( _err_res( $ctx, $sp, "contains matched $matches < minContains $minc" ) ) if( $matches < $minc );
-        return( _err_res( $ctx, $sp, "contains matched $matches > maxContains $maxc" ) ) if( $matches > $maxc );
+        return( _err_res( $ctx, $sp, "contains matched $matches < minContains $minc", 'minContains' ) ) if( $matches < $minc );
+        return( _err_res( $ctx, $sp, "contains matched $matches > maxContains $maxc", 'maxContains' ) ) if( $matches > $maxc );
     }
 
     if( exists( $S->{unevaluatedItems} ) )
@@ -1401,7 +1412,7 @@ sub _k_array_all
         my $UE = $S->{unevaluatedItems};
         if( !_is_true( $UE ) && !_is_hash( $UE ) )
         {
-            return( _err_res( $ctx, $sp, "unevaluatedItems not allowed at indices: " . join( ',', @unknown ) ) ) if( @unknown );
+            return( _err_res( $ctx, $sp, "unevaluatedItems not allowed at indices: " . join( ',', @unknown ), 'unevaluatedItems' ) ) if( @unknown );
         }
         elsif( ref( $UE ) eq 'HASH' )
         {
@@ -1441,7 +1452,7 @@ sub _k_combinator
             my $r = _v( $ctx, _join_ptr( $sp, "anyOf/$i" ), $S->{anyOf}->[ $i ], $inst );
             return( $r ) if( $r->{ok} );
         }
-        return( _err_res( $ctx, $sp, "instance does not satisfy anyOf" ) );
+        return( _err_res( $ctx, $sp, "instance does not satisfy anyOf", 'anyOf' ) );
     }
 
     if( $kw eq 'oneOf' )
@@ -1453,13 +1464,13 @@ sub _k_combinator
             push( @ok, $r ) if( $r->{ok} );
         }
         return( $ok[0] ) if( @ok == 1 );
-        return( _err_res( $ctx, $sp, "instance satisfies " . scalar( @ok ) . " schemas in oneOf (expected exactly 1)" ) );
+        return( _err_res( $ctx, $sp, "instance satisfies " . scalar( @ok ) . " schemas in oneOf (expected exactly 1)", 'oneOf' ) );
     }
 
     if( $kw eq 'not' )
     {
         my $r = _v( $ctx, _join_ptr( $sp, "not" ), $S->{not}, $inst );
-        return( _err_res( $ctx, $sp, "instance matches forbidden not-schema" ) ) if( $r->{ok} );
+        return( _err_res( $ctx, $sp, "instance matches forbidden not-schema", 'not' ) ) if( $r->{ok} );
         return( { ok => 1, props => {}, items => {} } );
     }
 
@@ -1470,7 +1481,7 @@ sub _k_const
 {
     my( $ctx, $v, $const, $ptr ) = @_;
     return(1) if( _json_equal( $v, $const ) );
-    return( _err( $ctx, $ptr, "const mismatch" ) );
+    return( _err( $ctx, $ptr, "const mismatch", 'const' ) );
 }
 
 sub _k_enum
@@ -1480,7 +1491,7 @@ sub _k_enum
     {
         return(1) if( _json_equal( $v, $e ) );
     }
-    return( _err( $ctx, $ptr, "value not in enum" ) );
+    return( _err( $ctx, $ptr, "value not in enum", 'enum' ) );
 }
 
 sub _k_format
@@ -1490,7 +1501,7 @@ sub _k_format
     return(1) unless( $cb );
     local $@;
     my $ok = eval{ $cb->( $s ) ? 1 : 0 };
-    return( $ok ? 1 : _err( $ctx, $ptr, "string fails format '$fmt'" ) );
+    return( $ok ? 1 : _err( $ctx, $ptr, "string fails format '$fmt'", 'format' ) );
 }
 
 sub _k_if_then_else
@@ -1521,7 +1532,7 @@ sub _k_number
         if( !defined( $arg ) || $arg <= 0 )
         {
             _t( $ctx, $ptr, 'multipleOf', undef, 'fail', 'arg<=0' ) if( $ctx->{trace_on} );
-            return( _err( $ctx, $ptr, "multipleOf must be > 0" ) );
+            return( _err( $ctx, $ptr, "multipleOf must be > 0", 'multipleOf' ) );
         }
         # Float-tolerant multiple check
         # my $ok = abs( ( $v / $arg ) - int( $v / $arg + 1e-10 ) ) < 1e-9;
@@ -1532,22 +1543,22 @@ sub _k_number
     elsif( $kw eq 'minimum' )
     {
         _t( $ctx, $ptr, 'minimum', undef, $v >= $arg ? 'pass' : 'fail', "$v >= $arg" ) if( $ctx->{trace_on} );
-        return( $v >= $arg ? 1 : _err( $ctx, $ptr, "number less than minimum $arg" ) );
+        return( $v >= $arg ? 1 : _err( $ctx, $ptr, "number less than minimum $arg", 'minimum' ) );
     }
     elsif( $kw eq 'maximum' )
     {
         _t( $ctx, $ptr, 'maximum', undef, $v <= $arg ? 'pass' : 'fail', "$v <= $arg" ) if( $ctx->{trace_on} );
-        return( $v <= $arg ? 1 : _err( $ctx, $ptr, "number greater than maximum $arg" ) );
+        return( $v <= $arg ? 1 : _err( $ctx, $ptr, "number greater than maximum $arg", 'maximum' ) );
     }
     elsif( $kw eq 'exclusiveMinimum' )
     {
         _t( $ctx, $ptr, 'exclusiveMinimum', undef, $v > $arg ? 'pass' : 'fail', "$v > $arg" ) if( $ctx->{trace_on} );
-        return( $v > $arg ? 1 : _err( $ctx, $ptr, "number not greater than exclusiveMinimum $arg" ) );
+        return( $v > $arg ? 1 : _err( $ctx, $ptr, "number not greater than exclusiveMinimum $arg", 'exclusiveMinimum' ) );
     }
     elsif( $kw eq 'exclusiveMaximum' )
     {
         _t( $ctx, $ptr, 'exclusiveMaximum', undef, $v < $arg ? 'pass' : 'fail', "$v < $arg" ) if( $ctx->{trace_on} );
-        return( $v < $arg ? 1 : _err( $ctx, $ptr, "number not less than exclusiveMaximum $arg" ) );
+        return( $v < $arg ? 1 : _err( $ctx, $ptr, "number not less than exclusiveMaximum $arg", 'exclusiveMaximum' ) );
     }
     return(1);
 }
@@ -1558,11 +1569,11 @@ sub _k_object_all
 
     if( exists( $S->{minProperties} ) && ( scalar( keys( %$H ) ) ) < $S->{minProperties} )
     {
-        return( _err_res( $ctx, $sp, "object has fewer than minProperties $S->{minProperties}" ) );
+        return( _err_res( $ctx, $sp, "object has fewer than minProperties $S->{minProperties}", 'minProperties' ) );
     }
     if( exists( $S->{maxProperties} ) && ( scalar( keys( %$H ) ) ) > $S->{maxProperties} )
     {
-        return( _err_res( $ctx, $sp, "object has more than maxProperties $S->{maxProperties}" ) );
+        return( _err_res( $ctx, $sp, "object has more than maxProperties $S->{maxProperties}", 'maxProperties' ) );
     }
 
     # Merge required from:
@@ -1597,7 +1608,7 @@ sub _k_object_all
     {
         next if( exists( $H->{ $rq } ) );
         _t( $ctx,$sp, 'required', undef, 'fail', $rq ) if( $ctx->{trace_on} );
-        return( _err_res( $ctx, _join_ptr( $sp, $rq ), "required property '$rq' is missing" ) );
+        return( _err_res( $ctx, _join_ptr( $sp, $rq ), "required property '$rq' is missing", 'required' ) );
     }
 
     if( exists( $S->{propertyNames} ) && ref( $S->{propertyNames} ) eq 'HASH' )
@@ -1647,7 +1658,7 @@ sub _k_object_all
         {
             if( $addl_set && !_is_true( $addl ) && !_is_hash( $addl ) )
             {
-                return( _err_res( $ctx, _join_ptr( $sp, $k ), "additionalProperties not allowed: '$k'" ) );
+                return( _err_res( $ctx, _join_ptr( $sp, $k ), "additionalProperties not allowed: '$k'", 'additionalProperties' ) );
             }
             elsif( ref( $addl ) eq 'HASH' )
             {
@@ -1666,7 +1677,7 @@ sub _k_object_all
             for my $need ( @{$depR->{ $k } || []} )
             {
                 next if( exists( $H->{ $need } ) );
-                return( _err_res( $ctx, _join_ptr( $sp, $need ), "dependentRequired: '$need' required when '$k' is present" ) );
+                return( _err_res( $ctx, _join_ptr( $sp, $need ), "dependentRequired: '$need' required when '$k' is present", 'dependentRequired' ) );
             }
         }
     }
@@ -1687,7 +1698,7 @@ sub _k_object_all
         my $UE = $S->{unevaluatedProperties};
         if( !_is_true( $UE ) && !_is_hash( $UE ) )
         {
-            return( _err_res( $ctx, $sp, "unevaluatedProperties not allowed: " . join( ',', @unknown ) ) ) if( @unknown );
+            return( _err_res( $ctx, $sp, "unevaluatedProperties not allowed: " . join( ',', @unknown ), 'unevaluatedProperties' ) ) if( @unknown );
         }
         elsif( ref( $UE ) eq 'HASH' )
         {
@@ -1711,14 +1722,14 @@ sub _k_string
     if( exists( $S->{minLength} ) && $len < $S->{minLength} )
     {
         _t( $ctx, $ptr, 'minLength', undef, 'fail', "len=$len < $S->{minLength}" ) if( $ctx->{trace_on} );
-        return( _err( $ctx, $ptr, "string shorter than minLength $S->{minLength}" ) );
+        return( _err( $ctx, $ptr, "string shorter than minLength $S->{minLength}", 'minLength' ) );
     }
     _t( $ctx, $ptr, 'minLength', undef, 'pass', "len=$len >= $S->{minLength}" ) if( exists( $S->{minLength} ) && $ctx->{trace_on} );
 
     if( exists( $S->{maxLength} ) && $len > $S->{maxLength} )
     {
         _t( $ctx, $ptr, 'maxLength', undef, 'fail', "len=$len > $S->{maxLength}" ) if( $ctx->{trace_on} );
-        return( _err( $ctx, $ptr, "string longer than maxLength $S->{maxLength}" ) );
+        return( _err( $ctx, $ptr, "string longer than maxLength $S->{maxLength}", 'maxLength' ) );
     }
     _t( $ctx, $ptr, 'maxLength', undef, 'pass', "len=$len <= $S->{maxLength}" ) if( exists( $S->{maxLength} ) && $ctx->{trace_on} );
 
@@ -1728,7 +1739,7 @@ sub _k_string
         local $@;
         my $ok = eval{ $s =~ /$re/ };
         _t( $ctx, $ptr, 'pattern', undef, $ok ? 'pass' : 'fail', "/$re/" ) if( $ctx->{trace_on} );
-        return( _err( $ctx, $ptr, "string does not match pattern /$re/" ) ) unless( $ok );
+        return( _err( $ctx, $ptr, "string does not match pattern /$re/", 'pattern' ) ) unless( $ok );
     }
     return(1);
 }
@@ -1760,7 +1771,7 @@ sub _k_type
     }
 
     my $exp = join( '|', map{ ref( $_ ) ? 'schema' : $_ } @alts );
-    return( _err( $ctx, $ptr, "type mismatch: expected $exp" ) );
+    return( _err( $ctx, $ptr, "type mismatch: expected $exp", 'type' ) );
 }
 
 sub _match_type
@@ -2093,7 +2104,7 @@ sub _v_node
                 my $dec = _content_decode( $ctx, $schema->{contentEncoding}, $bytes );
                 if( !defined( $dec ) )
                 {
-                    return( _err_res( $ctx, $ptr, "contentEncoding '$schema->{contentEncoding}' decode failed" ) ) if( $assert );
+                    return( _err_res( $ctx, $ptr, "contentEncoding '$schema->{contentEncoding}' decode failed", 'contentEncoding' ) ) if( $assert );
                 }
                 else
                 {
@@ -2109,7 +2120,7 @@ sub _v_node
                     my( $ok, $msg, $maybe_decoded ) = _safe_invoke( $cb, $bytes, $params );
                     if( !$ok )
                     {
-                        return( _err_res( $ctx, $ptr, ( $msg || "contentMediaType '$mt' validation failed" ) ) ) if( $assert );
+                        return( _err_res( $ctx, $ptr, ( $msg || "contentMediaType '$mt' validation failed", 'contentMediaType' ) ) ) if( $assert );
                     }
                     $decoded_ref = $maybe_decoded if( ref( $maybe_decoded ) );
                     $bytes = $maybe_decoded if( defined( $maybe_decoded ) && !ref( $maybe_decoded ) );
@@ -2127,7 +2138,7 @@ sub _v_node
                         } ? 1 : 0;
                         if( !$ok && $assert )
                         {
-                            return( _err_res( $ctx, $ptr, "contentMediaType '$mt' invalid UTF-8" ) );
+                            return( _err_res( $ctx, $ptr, "contentMediaType '$mt' invalid UTF-8", 'contentMediaType' ) );
                         }
                     }
                 }
@@ -2148,7 +2159,7 @@ sub _v_node
 
                 if( !defined( $val ) )
                 {
-                    return( _err_res( $ctx, $ptr, "contentSchema present but payload not JSON-decodable" ) ) if( $assert );
+                    return( _err_res( $ctx, $ptr, "contentSchema present but payload not JSON-decodable", 'contentSchema' ) ) if( $assert );
                 }
                 else
                 {
@@ -2240,16 +2251,49 @@ use warnings;
 
 sub new
 {
-    my( $that, $path, $mesg ) = @_;
-    return( bless({ path => $path, message => $mesg } => ( ref( $that ) || $that ) ) );
+    my $that = shift( @_ );
+    my $ref = {};
+    # Legacy instantiation
+    # We make sure this is not one single option that was provided to us.
+    if( @_ == 2 && $_[0] !~ /^(?:path|message|keyword|schema_pointer)$/ )
+    {
+        @$ref{qw( path message )} = @_;
+    }
+    else
+    {
+        my $args = { @_ };
+        for( qw( path message keyword schema_pointer ) )
+        {
+            $ref->{ $_ } = $args->{ $_ } if( exists( $args->{ $_ } ) );
+        }
+    }
+    return( bless( $ref => ( ref( $that ) || $that ) ) );
+}
+
+sub as_hash
+{
+    my $self = shift( @_ );
+    my $ref = {};
+    my @keys = qw( keyword message path schema_pointer );
+    @$ref{ @keys } = @$self{ @keys };
+    return( $ref );
 }
 
 sub as_string
 {
     my $self = shift( @_ );
+    my $sp   = $self->schema_pointer // '';
     my $path = $self->path // '';
     my $msg  = $self->message // '';
-    return( "${path}: ${msg}" );
+    # Dual-path if avail
+    return( $sp ? "${sp} → ${path}: ${msg}" : "${path}: ${msg}" );
+}
+
+sub keyword
+{
+    my $self = shift( @_ );
+    $self->{keyword} = shift( @_ ) if( @_ );
+    return( $self->{keyword} );
 }
 
 sub message
@@ -2266,6 +2310,13 @@ sub path
     return( $self->{path} );
 }
 
+sub schema_pointer
+{
+    my $self = shift( @_ );
+    $self->{schema_pointer} = shift( @_ ) if( @_ );
+    return( $self->{schema_pointer} );
+}
+
 sub _obj_eq
 {
     no overloading;
@@ -2339,7 +2390,7 @@ JSON::Schema::Validate - Lean, recursion-safe JSON Schema validator (Draft 2020-
 
 =head1 VERSION
 
-v0.2.0
+v0.3.0
 
 =head1 DESCRIPTION
 
